@@ -18,7 +18,7 @@ class UserState(CacheState[UserStateType]):
 
     @property
     def _state_path(self) -> str:
-        return f'{super()._cache_dir}/user-state.json'
+        return f'{self._user_cache_dir}/user-state.json'
 
     def _get_default_state(self) -> UserStateType:
         return UserStateType(refresh_token=None)
