@@ -1,12 +1,12 @@
-from abc import ABC, abstractmethod
 import io
-import time
+import math
+import multiprocessing
+from abc import ABC, abstractmethod
+from multiprocessing.pool import ThreadPool
+from typing import Callable, Optional
 
-import requests
-from requests import Response
-
-from biolib.biolib_logging import logger
-from biolib.typing_utils import Optional, Dict
+from biolib._internal.http_client import HttpClient
+from biolib.typing_utils import Iterator
 
 
 class RemoteEndpoint(ABC):
@@ -16,7 +16,6 @@ class RemoteEndpoint(ABC):
 
 
 class IndexableBuffer(ABC):
-
     def __init__(self):
         self.pointer = 0
 
@@ -46,48 +45,28 @@ class IndexableBuffer(ABC):
         return data
 
 
-def _get_request_with_retries_for_timeout(
-        url: str,
-        headers: Optional[Dict[str, str]] = None,
-        stream: bool = False,
-) -> Response:
-    retries = 10
-    for retry_count in range(retries):
-        if retry_count > 0:
-            time.sleep(1 * retry_count)
-            logger.debug('Retrying HTTP GET request...')
+class LocalFileIndexableBuffer(IndexableBuffer):
+    def __init__(self, filename: str):
+        super().__init__()
+        self._filehandle = open(filename, 'rb')
 
-        try:
-            response = requests.get(
-                headers=headers,
-                stream=stream,
-                timeout=10,
-                url=url,
-            )
-        except requests.exceptions.ReadTimeout:
-            logger.debug('Got read timeout of HTTP GET request')
-            continue
+    def get_data(self, start: int, length: int) -> bytes:
+        if length < 0:
+            raise Exception('get_data length must be positive')
 
-        if response.status_code == 502:
-            logger.debug('Got 502 status for HTTP GET')
-            continue
+        if length == 0:
+            return bytes(0)
 
-        if response.status_code == 400:
-            logger.error(
-                f'GET request for "{url}" got response with status {response.status_code} and '
-                f'content: {response.content.decode()}'
-            )
+        self._filehandle.seek(start)
+        data: bytes = self._filehandle.read(length)
 
-        if not response.ok:
-            raise Exception(f'GET request for "{url}" got response with status {response.status_code}')
+        if len(data) != length:
+            raise Exception(f'get_data got response of unexpected length. Got {len(data)} expected {length}.')
 
-        return response
-
-    raise Exception(f'Failed to complete HTTP GET request for "{url}" after {retries} retries')
+        return data
 
 
 class RemoteIndexableBuffer(IndexableBuffer):
-
     def __init__(self, endpoint: RemoteEndpoint):
         super().__init__()
         self._endpoint = endpoint
@@ -100,7 +79,7 @@ class RemoteIndexableBuffer(IndexableBuffer):
             return bytes(0)
 
         end = start + length - 1
-        response = _get_request_with_retries_for_timeout(
+        response = HttpClient.request(
             url=self._endpoint.get_remote_url(),
             headers={'range': f'bytes={start}-{end}'},
         )
@@ -113,7 +92,6 @@ class RemoteIndexableBuffer(IndexableBuffer):
 
 
 class InMemoryIndexableBuffer(IndexableBuffer):
-
     def __init__(self, data: bytes):
         super().__init__()
         self._buffer = data
@@ -128,11 +106,18 @@ class InMemoryIndexableBuffer(IndexableBuffer):
 
 
 class LazyLoadedFile:
-
-    def __init__(self, path: str, buffer: IndexableBuffer, start: int, length: int):
+    def __init__(
+        self,
+        path: str,
+        buffer: IndexableBuffer,
+        start: Optional[int],
+        length: int,
+        start_func: Optional[Callable[[], int]] = None,
+    ):
         self._path = path
         self._buffer = buffer
         self._start = start
+        self._start_func = start_func
         self._length = length
 
     def __repr__(self) -> str:
@@ -142,8 +127,16 @@ class LazyLoadedFile:
     def path(self) -> str:
         return self._path
 
+    @property
+    def name(self) -> str:
+        return self._path.split('/')[-1]
+
     @property
     def start(self) -> int:
+        if self._start is None:
+            assert self._start_func is not None, 'No start function or start value'
+            self._start = self._start_func()
+
         return self._start
 
     @property
@@ -153,5 +146,30 @@ class LazyLoadedFile:
     def get_file_handle(self) -> io.BufferedIOBase:
         return io.BytesIO(self.get_data())
 
-    def get_data(self) -> bytes:
-        return self._buffer.get_data(start=self._start, length=self._length)
+    def get_data(self, start=0, length=None) -> bytes:
+        start_offset = start + self.start
+        # make sure length doesn't go outside file boundaries
+        length_to_end_of_file = max(self._length - start, 0)
+        length_to_request = length_to_end_of_file if length is None else min(length, length_to_end_of_file)
+        return self._buffer.get_data(start=start_offset, length=length_to_request)
+
+    def get_data_iterator(self) -> Iterator[bytes]:
+        if self._length == 0:
+            yield b''
+        else:
+            chunk_size = 50_000_000
+            chunk_count = math.ceil(self._length / chunk_size)
+            chunk_indices_iterator = range(chunk_count - 1)
+
+            def get_chunk(chunk_index: int) -> bytes:
+                return self._buffer.get_data(start=self.start + chunk_index * chunk_size, length=chunk_size)
+
+            if chunk_count > 1:
+                with ThreadPool(processes=min(16, chunk_count, multiprocessing.cpu_count() - 1)) as pool:
+                    yield from pool.imap(func=get_chunk, iterable=chunk_indices_iterator)
+
+            data_already_yielded = (chunk_count - 1) * chunk_size
+            yield self._buffer.get_data(
+                start=self.start + data_already_yielded,
+                length=self._length - data_already_yielded,
+            )
