@@ -1,11 +1,11 @@
-from docker.models.networks import Network  # type: ignore
+from docker.models.networks import Network
 
-from biolib.compute_node.job_worker.large_file_system import LargeFileSystem
-from biolib.compute_node.webserver.webserver_types import ComputeNodeInfo
-from biolib.typing_utils import TypedDict, Callable, Optional, List, Dict
-from biolib.compute_node.remote_host_proxy import RemoteHostProxy
 from biolib.biolib_api_client.app_types import Module
-from biolib.biolib_api_client.job_types import CreatedJobDict, CloudJob
+from biolib.biolib_api_client.job_types import CloudJob, CreatedJobDict
+from biolib.compute_node.job_worker.large_file_system import LargeFileSystem
+from biolib.compute_node.remote_host_proxy import RemoteHostProxy
+from biolib.compute_node.webserver.webserver_types import ComputeNodeInfo
+from biolib.typing_utils import Callable, Dict, List, Optional, TypedDict
 
 
 class StatusUpdate(TypedDict):
@@ -43,6 +43,7 @@ class LocalExecutorOptions(TypedDict):
     send_system_exception: SendSystemExceptionType
     send_stdout_and_stderr: SendStdoutAndStderrType
 
+
 class MetadataToSaveOutput(TypedDict):
     arguments: List[str]
     startup_error_string: Optional[str]
