@@ -1,44 +1,56 @@
 import io
 import json
-import socket
+import multiprocessing
+import os
 import shlex
+import signal
+import socket
 import sys
 import tempfile
 import zipfile
-from time import time
 from queue import Queue
-import multiprocessing
-import os
-import signal
+from time import time
 from types import FrameType
 
 from docker.models.networks import Network  # type: ignore
+from docker.types import IPAMConfig, IPAMPool
 
-from biolib._internal.http_client import HttpClient
-from biolib.biolib_binary_format.stdout_and_stderr import StdoutAndStderr
-from biolib.compute_node.job_worker.job_legacy_input_wait_timeout_thread import JobLegacyInputWaitTimeout
-from biolib.compute_node.job_worker.job_storage import JobStorage
-from biolib.compute_node.job_worker.large_file_system import LargeFileSystem
-from biolib.biolib_errors import DockerContainerNotFoundDuringExecutionException, BioLibError, \
-    StorageDownloadFailed
-from biolib.compute_node.job_worker.job_max_runtime_timer_thread import JobMaxRuntimeTimerThread
-from biolib.compute_node.remote_host_proxy import RemoteHostProxy
-from biolib.typing_utils import Optional, List, Dict
 from biolib import utils
-from biolib.biolib_api_client import ModuleEnvironment, CreatedJobDict, JobWrapper, Module, AppVersionOnJob, \
-    BiolibApiClient, RemoteHost
+from biolib._internal.http_client import HttpClient
+from biolib.biolib_api_client import (
+    AppVersionOnJob,
+    BiolibApiClient,
+    CreatedJobDict,
+    JobWrapper,
+    Module,
+    ModuleEnvironment,
+)
 from biolib.biolib_api_client.biolib_job_api import BiolibJobApi
+from biolib.biolib_binary_format import (
+    InMemoryIndexableBuffer,
+    ModuleInput,
+    ModuleOutputV2,
+    SavedJob,
+    SystemException,
+    SystemStatusUpdate,
+)
+from biolib.biolib_binary_format.stdout_and_stderr import StdoutAndStderr
 from biolib.biolib_docker_client import BiolibDockerClient
+from biolib.biolib_errors import BioLibError, DockerContainerNotFoundDuringExecutionException, StorageDownloadFailed
 from biolib.biolib_logging import logger, logger_no_user_data
 from biolib.compute_node.job_worker.executors import DockerExecutor
 from biolib.compute_node.job_worker.executors.types import LocalExecutorOptions, StatusUpdate
-from biolib.compute_node.socker_listener_thread import SocketListenerThread
-from biolib.compute_node.socket_sender_thread import SocketSenderThread
+from biolib.compute_node.job_worker.job_legacy_input_wait_timeout_thread import JobLegacyInputWaitTimeout
+from biolib.compute_node.job_worker.job_max_runtime_timer_thread import JobMaxRuntimeTimerThread
+from biolib.compute_node.job_worker.job_storage import JobStorage
+from biolib.compute_node.job_worker.large_file_system import LargeFileSystem
 from biolib.compute_node.job_worker.mappings import Mappings, path_without_first_folder
 from biolib.compute_node.job_worker.utils import ComputeProcessException, log_disk_and_memory_usage_info
-from biolib.compute_node.utils import get_package_type, SystemExceptionCodes, SystemExceptionCodeMap
-from biolib.biolib_binary_format import SavedJob, SystemStatusUpdate, ModuleInput, SystemException, \
-    ModuleOutputV2, InMemoryIndexableBuffer
+from biolib.compute_node.remote_host_proxy import RemoteHostMapping, RemoteHostProxy, get_static_ip_from_network
+from biolib.compute_node.socker_listener_thread import SocketListenerThread
+from biolib.compute_node.socket_sender_thread import SocketSenderThread
+from biolib.compute_node.utils import SystemExceptionCodeMap, SystemExceptionCodes, get_package_type
+from biolib.typing_utils import Dict, List, Optional
 
 SOCKET_HOST = '127.0.0.1'
 
@@ -199,6 +211,7 @@ class JobWorker:
             executor.cleanup()
 
         proxy_count = len(self._remote_host_proxies)
+        cleaned_networks = set()
         if proxy_count > 0:
             logger_no_user_data.debug('Cleaning up proxies...')
             proxy_cleanup_start_time = time()
@@ -210,11 +223,20 @@ class JobWorker:
                     logger_no_user_data.error('Failed to clean up remote host proxy')
                     logger.error(exception)
 
+                for network in proxy.get_remote_host_networks():
+                    try:
+                        self._cleanup_network(network)
+                        cleaned_networks.add(network.id)
+                    except Exception as exception:  # pylint: disable=broad-except
+                        logger_no_user_data.error(f'Failed to clean up network {network.name}')
+                        logger.error(exception)
+
             self._remote_host_proxies = []
             logger_no_user_data.debug(f'Cleaned up {proxy_count} proxies in {time() - proxy_cleanup_start_time}')
 
         logger_no_user_data.debug('Cleaning up networks...')
-        self._cleanup_network(self._internal_network)
+        if self._internal_network and self._internal_network.id not in cleaned_networks:
+            self._cleanup_network(self._internal_network)
         self._internal_network = None
         logger_no_user_data.debug('Cleaned up networks...')
 
@@ -265,13 +287,6 @@ class JobWorker:
         app_version = job['app_version']
         job_id = job['public_id']
         remote_hosts = app_version['remote_hosts']
-        if utils.IS_RUNNING_IN_CLOUD:
-            remote_hosts.append(
-                {
-                    'hostname': 'AppCallerProxy',
-                },
-            )
-
         docker_client = BiolibDockerClient.get_docker_client()
         try:
             self._internal_network = docker_client.networks.create(
@@ -288,7 +303,7 @@ class JobWorker:
             ) from exception
 
         if len(remote_hosts) > 0:
-            logger_no_user_data.debug(f'Job "{job_id}" starting proxies for remote hosts: {remote_hosts}')
+            logger_no_user_data.debug(f'Job "{job_id}" starting proxy for remote hosts: {remote_hosts}')
             try:
                 hostname_to_ports: Dict[str, List[int]] = {}
                 for remote_host in remote_hosts:
@@ -304,15 +319,44 @@ class JobWorker:
                     else:
                         hostname_to_ports[hostname] = [port]
 
+                remote_host_mappings: List[RemoteHostMapping] = []
                 for hostname, ports in hostname_to_ports.items():
-                    remote_host_proxy = RemoteHostProxy(
-                        remote_host=RemoteHost(hostname=hostname),
-                        internal_network=self._internal_network,
-                        job_id=job_id,
+                    network_index = len(remote_host_mappings)
+                    subnet = f'172.28.{network_index}.0/24'
+
+                    ipam_pool = IPAMPool(subnet=subnet)
+                    ipam_config = IPAMConfig(pool_configs=[ipam_pool])
+
+                    network = docker_client.networks.create(
+                        name=f'biolib-remote-host-network-{job_id}-{network_index}',
+                        internal=True,
+                        driver='bridge',
+                        ipam=ipam_config,
+                    )
+
+                    static_ip = get_static_ip_from_network(network, offset=2)
+
+                    mapping = RemoteHostMapping(
+                        hostname=hostname,
                         ports=ports,
+                        network=network,
+                        static_ip=static_ip,
+                    )
+                    remote_host_mappings.append(mapping)
+                    logger_no_user_data.debug(
+                        f'Created network {network.name} with static IP {static_ip} for host {hostname}'
+                    )
+
+                if remote_host_mappings:
+                    remote_host_proxy = RemoteHostProxy(
+                        remote_host_mappings=remote_host_mappings,
+                        job_id=job_id,
+                        app_caller_network=None,
                     )
                     remote_host_proxy.start()
                     self._remote_host_proxies.append(remote_host_proxy)
+                    num_hosts = len(remote_host_mappings)
+                    logger_no_user_data.debug(f'Started single proxy container for {num_hosts} remote hosts')
 
             except Exception as exception:
                 raise ComputeProcessException(
@@ -322,7 +366,23 @@ class JobWorker:
                     may_contain_user_data=False
                 ) from exception
 
-            logger_no_user_data.debug(f'Job "{job_id}" startup of remote host proxies completed')
+        if utils.IS_RUNNING_IN_CLOUD:
+            try:
+                app_caller_proxy = RemoteHostProxy(
+                    remote_host_mappings=[],
+                    job_id=job_id,
+                    app_caller_network=self._internal_network,
+                )
+                app_caller_proxy.start()
+                self._remote_host_proxies.append(app_caller_proxy)
+                logger_no_user_data.debug('Started app caller proxy')
+            except Exception as exception:
+                raise ComputeProcessException(
+                    exception,
+                    SystemExceptionCodes.FAILED_TO_START_REMOTE_HOST_PROXIES.value,
+                    self.send_system_exception,
+                    may_contain_user_data=False
+                ) from exception
 
     def _run_app_version(
             self,
